#include "dbg.h"
#include "stdafx.h"

#define TF_BIT 0x100
//undo the macro that windows.h has defined for us
#undef GetCurrentDirectory

public ref struct DbgApp
{
public:
	bool bcrashed;
	String ^crashhash;
};

String ^GetCrashHash(HANDLE hThread);
DbgApp ^LaunchProg(String ^strCmdLine, int intTimeout);



DbgApp ^LaunchProg(String ^strCmdLine, int intTimeout)
{
	DbgApp ^retStatus = gcnew DbgApp;
	retStatus->bcrashed = false;
	retStatus->crashhash = "";

	TCHAR *wszCmdLine;

	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	DEBUG_EVENT dbgEvent;
	DWORD dbgStatus = DBG_CONTINUE;
	bool bHitBreak = false;
	HANDLE hThread;
	DateTime dtInit;
	DWORD dwLastExAddr = -1;
	DateTime dtTimeout;
	bool isRunning;
	int iTimeout;

	
	wszCmdLine = (TCHAR*)(void*)Marshal::StringToHGlobalUni(strCmdLine);

	//create the startupinfo structure
	ZeroMemory( &si, sizeof(STARTUPINFO) );
	si.cb = sizeof(STARTUPINFO); 
	si.wShowWindow = SW_SHOW;

	if (CreateProcess(0, wszCmdLine, 0, 0, false, DEBUG_PROCESS, 0, 0, &si, &pi))
	{
		dbgStatus = DBG_CONTINUE;
		isRunning = TRUE;
		iTimeout = INFINITE;
		bHitBreak = false;
		//we don't set up the timeout timer until we hit the first breakpoint

		EXCEPTION_DEBUG_INFO dbgExInfo;
		while (isRunning)
		{
			WaitForDebugEvent(&dbgEvent, iTimeout * 1000); //iTimeout is in seconds
			
			//check if we've timed out, unless the program hasnt officially started yet
			if (iTimeout!=INFINITE)
				if (DateTime::Compare(DateTime::Now, dtTimeout) >= 0)
				{
					//we timed out
					retStatus->bcrashed = false;
					TerminateProcess(pi.hProcess, 0);
				}

			hThread = OpenThread(THREAD_ALL_ACCESS, FALSE, dbgEvent.dwThreadId);
			
			switch (dbgEvent.dwDebugEventCode)
			{
			default:
				dbgStatus = DBG_CONTINUE;
				break;
			case EXCEPTION_DEBUG_EVENT:

				dbgExInfo = (EXCEPTION_DEBUG_INFO)dbgEvent.u.Exception;

				switch (dbgEvent.u.Exception.ExceptionRecord.ExceptionCode) 
				{

				case EXCEPTION_BREAKPOINT:
					if (!bHitBreak)	//continue past the frist breakpoint
					{
						//intTimeout = 0 means no time limit
						if (intTimeout == 0)
							iTimeout = INFINITE;
						else
						{
							//set up the timeout variables
							iTimeout = intTimeout;
							dtInit = DateTime::Now;
							dtTimeout = dtInit.AddSeconds(intTimeout);
						}
						dbgStatus = DBG_CONTINUE;
						bHitBreak = true;
					}
					else{
						//we are definately where we shouldnt be
						retStatus->bcrashed = true;
						isRunning = false;
						dbgStatus = DBG_EXCEPTION_NOT_HANDLED;
						TerminateProcess(pi.hProcess, 0);
					}
					break;

				default:
					//unrecognized event... could be a crash
					//if (dbgEvent.u.Exception.dwFirstChance)
					if (dbgExInfo.ExceptionRecord.ExceptionFlags == EXCEPTION_NONCONTINUABLE || dwLastExAddr == (DWORD)dbgExInfo.ExceptionRecord.ExceptionAddress)
					{
						//second chance exception. terminate the process
						retStatus->bcrashed = TRUE;
						retStatus->crashhash = GetCrashHash(hThread);
						dbgStatus = DBG_EXCEPTION_NOT_HANDLED;
						TerminateProcess(pi.hProcess, 0);
						//dont set isRunning to false yet... EXIT_PROCESS_DEBUG_EVENT will still be called
					}
					else
					{
						//first chance exception. save the address for later
						dwLastExAddr = (DWORD)dbgExInfo.ExceptionRecord.ExceptionAddress;
						dbgStatus = DBG_CONTINUE;
					}
					break;
				} 

				break;
			case EXIT_PROCESS_DEBUG_EVENT:
				//our process exited
				isRunning = false;
				dbgStatus = DBG_CONTINUE;
				break;
			}
			//send the process the debug signal
			ContinueDebugEvent(dbgEvent.dwProcessId, dbgEvent.dwThreadId, dbgStatus);
			dbgStatus = DBG_CONTINUE;
		}
	}

	
	//clean up
	if (hThread)
		CloseHandle(hThread);
	
	Marshal::FreeHGlobal((IntPtr)wszCmdLine);
	return retStatus;
}

String ^GetCrashHash(HANDLE hThread){
	//this function is supposed to return a unique hash for our crash
	//right now i just use EIP
	//if anyone has any ideas as to a better way please let me know
	CONTEXT dbgContext;
	TCHAR wszEIP[9];
	String ^strRet;

	//get all the information
	dbgContext.ContextFlags = CONTEXT_ALL;
	GetThreadContext(hThread, &dbgContext);

	//save eip
	swprintf(wszEIP, 9, L"%.8x", dbgContext.Eip);
	strRet = Marshal::PtrToStringAuto((IntPtr)wszEIP);

	return strRet;
}

array<String ^> ^GetURLs(){
	//open the url list file
	FileStream ^fs = File::Open(Directory::GetCurrentDirectory() + "\\urls",FileMode::Open);
	String ^strContents;
	StreamReader ^sr = gcnew StreamReader(fs);
	array<String ^> ^mngdURLs = {};

	//pouplate the managed array, leaving out comments and blank lines
	strContents = sr->ReadLine()->Trim();
	while(strContents){
		if (strContents->Length)
			if (strContents[0]!='#'){
				Array::Resize(mngdURLs,mngdURLs->Length+1);
				mngdURLs[mngdURLs->Length-1] = strContents;
			}
		strContents = sr->ReadLine();
	}
	return mngdURLs;
}

void showUsage(){
	printf("Usage:\r\nautofuzz.exe [/to timeout] [/f fuzzer(s)] /p binary path");
}

int main(array<System::String ^> ^args)
{
	DbgApp ^dbg;

	String ^strAppPath;
	String ^strFuzzPath;
	String ^strURL;
	String ^strFuzzSettings;
	String ^strUrlParam;
	String ^fuzzbasefilename;
	String ^crshdir = Directory::GetCurrentDirectory() + "\\crashes";
	FileStream ^crshlog;
	StreamWriter ^sw;
	int iSeed;
	int iPer;

	//create the crashes directory
	if (!Directory::Exists(crshdir))
		Directory::CreateDirectory(crshdir);

	array<String ^> ^URLs = GetURLs();
	array<String ^> ^TempStrArray;
	array<String ^> ^Fuzzers;
	Random ^rnd = gcnew Random();

	int iTimeout = 20;	//1 minute default timeout

	//get parameters
	for (int a=0;a<args->Length;a++)
	{
		if(args[a] == "/t")
		{
			try{
				a++;
				if ((a<args->Length) && (Convert::ToInt32(args[a])>=0))
				{
					iTimeout = Convert::ToInt32(args[a]);
					continue;
				}
			}catch(...){}
			//bad argument
			showUsage();
			return 0;
		}
		if(args[a] == "/p")
		{
			try{
				a++;
				if (a<args->Length)
				{
					strAppPath = args[a];
					continue;
				}
			}catch(...){}
			//bad argument
			showUsage();
			return 0;
		}
		if(args[a] == "/f")
		{
			try{
				a++;
				if (a<args->Length)
				{
					Fuzzers = args[a]->Split(',');
					continue;
				}
			}catch(...){}
			//bad argument
			showUsage();
			return 0;
		}
		if (args[a] == "?")
		{
			showUsage();
			return 0;
		}
	}
	if (!strAppPath)
	{
		showUsage();
		return 0;
	}

	//see if we havent specified a fuzzer yet
	if (!Fuzzers)
		Fuzzers = Directory::GetFiles(Directory::GetCurrentDirectory() + "\\fuzzers\\","*.js");

	while(1)
	{
		//choose a fuzzer
		TempStrArray = Fuzzers[rnd->Next(Fuzzers->GetUpperBound(0))]->Split('\\');
		strFuzzPath = TempStrArray[TempStrArray->GetUpperBound(0)];

		//choose an url
		strURL = URLs[rnd->Next(URLs->Length-1)];

		//choose the fuzzer settings
		iSeed = rnd->Next(1000000);
		iPer = rnd->Next(15) * rnd->Next(15) + 5;
		strFuzzSettings = Convert::ToString(iSeed) + ",0," + Convert::ToString(iPer) + "," + Convert::ToString(iPer);

		//build the url parameter
		strUrlParam = strURL + "#squarefree-af!" + strFuzzPath + "!" + strFuzzSettings;

		//launch the program
		dbg = LaunchProg("\"" + strAppPath + "\" \"" + strUrlParam + "\"", iTimeout);
		if (dbg->bcrashed)
		{
			//hell yeah!

			//log everything to crashes\*fuzzername*\*crashhash*\log.txt
			fuzzbasefilename = strFuzzPath->Split('.')[0];
			if (!Directory::Exists(crshdir + "\\" + fuzzbasefilename))
				Directory::CreateDirectory(crshdir + "\\" + fuzzbasefilename);
			if (!Directory::Exists(crshdir + "\\" + fuzzbasefilename + "\\" + dbg->crashhash))
				Directory::CreateDirectory(crshdir + "\\" + fuzzbasefilename + "\\" + dbg->crashhash);
			crshlog = File::Open(crshdir + "\\" + fuzzbasefilename + "\\" + dbg->crashhash + "\\log.txt",FileMode::OpenOrCreate);
			sw = gcnew StreamWriter(crshlog);
			sw->BaseStream->Seek(0, SeekOrigin::End);
			sw->WriteLine(strUrlParam);
			sw->Flush();
			sw->Close();
			crshlog->Close();
		}
	}
	//clean up
    return 0;
}
//wow... 150 lines for af_timed_run and over 320 lines for autofuzz.cpp. i need to learn python ;)