// === Paste output from first run here, replacing the following two lines: ===

var fuzzSettings = null;
var fuzzCommands = null;

// === End pasteover ===


var fuzzerName = "DOM Range and designMode fuzzer 3.1";
var recordStrategy = "Record as it goes";
var suggestedMaxSteps = 5000;


var targetWindow;
var targetDocument;



function offsets(n) 
{ 
  if (n) {
    if (n.nodeType == Node.TEXT_NODE) 
      return n.data.length + 1; 
    else 
      return n.childNodes.length + 1; 
  }
  else
    return 1;
}
      

function makeCommand() 
{
  return (rndElt(commandMakers))();
}




function makePartialCommandRange()
{ 
  var n1 = rnd(allNodes.length);
  var n2 = rnd(allNodes.length);
  
  return "var r = targetDocument.createRange(); " +
         "r.setStart(allNodes[" + n1 + "], " + rnd(offsets(allNodes[n1])) + "); "  +
         "r.setEnd(allNodes[" + n2 + "], " + rnd(offsets(allNodes[n2])) + "); ";
}


var commandMakers = [

  function () {
    // Add a random range to the selection
    return makePartialCommandRange() + "targetWindow.getSelection().addRange(r);";
  },
   
  function () {
    // Add a random range to allRanges
    return makePartialCommandRange() + "allRanges[" + allRanges.length + "] = r;";
  },
   
  function () {
    // Add a random range to allRanges
    return "allRanges[" + allRanges.length + "] = targetWindow.getSelection().getRangeAt(0);";
  },
   
  function () {
    // Delete range
    if (allRanges.length)
      return "allRanges[" + rnd(allRanges.length) + "].deleteContents()";
    else
      return "";
  },

/*
  function () {
    return "allNodes[" + rnd(allNodes.length) + "].innerHTML += allNodes[" + rnd(allNodes.length) + "].innerHTML;";
  },
*/

/*
  function () {
    // The null forces the baseline to be constant while reducing.
    return "allNodes[" + allNodes.length + "] = null; addElements();";
  },
*/

/*
  function () {
    return "allFragments[" + allFragments.length + "] = allRanges[" + rnd(allRanges.length) + "].extractContents();";
  },
*/

/*
  function () {
    return "allRanges[" + rnd(allRanges.length) + "].deleteContents();";
  },

  function () {
    return "allNodes[" + rnd(allNodes.length) + "].appendChild(allFragments[" + rnd(allFragments.length) + ");";
  },
*/

  // Strange range features

/*
  function () {
    return "allRanges[" + rnd(allRanges.length) + "].insertNode(allNodes[" + rnd(allNodes.length) + "]);";
  },
*/

/*

  function () {
    return "allRanges[" + rnd(allRanges.length) + "].insertNode(allFragments[" + rnd(allFragments.length) + "]);";
  },
*/

/*
  function () {
    return "allRanges[" + rnd(allRanges.length) + "].surroundContents(allNodes[" + rnd(allNodes.length) + "]);";
  },
*/


  function () {
    return "allRanges[" + rnd(allRanges.length) + "].surroundContents(targetDocument.createElementNS('http://www.w3.org/1999/xhtml', 'i'));";
  },


/*
  function () {
    return "allRanges[" + rnd(allRanges.length) + "].insertNode(targetDocument.createElementNS('http://www.w3.org/1999/xhtml', 'hr'));";
  },
*/

    

  function () {
    return "allRanges[" + rnd(allRanges.length) + "].detach();";
  },

  function () {
    return "allRanges[" + rnd(allRanges.length) + "].compareBoundaryPoints(Range.START_TO_START, allRanges[" + rnd(allRanges.length) + "]);";
  },

  function () {
    return "allRanges[" + rnd(allRanges.length) + "].commonAncestorContainer;";
  },


/*   
  function () {
    // Delete selection
    return "targetWindow.getSelection().deleteFromDocument()";
  },
*/  

  function () {
    // Deselect
    return "targetWindow.getSelection().removeAllRanges()";
  },
  
/*
  function () {
    // From StirDOM: appendChild.
    return "allNodes[" + rnd(allNodes.length) + "].appendChild(allNodes[" + rnd(allNodes.length) + "]);";
  },
*/
  

/*
  function () {
       return "targetDocument.execCommand('" + rndElt(["bold", "italic", "justifycenter", "justifyleft", "removeformat", "undo", "redo", "indent", "outdent"]) + "', false, null)";
  }
*/

               
];
  
  
function setTextContent(n, s)
{
  // Safari sucks, so we can't use foo.textContent.
  while (n.firstChild)
    n.removeChild(n.firstChild);
    
  n.appendChild(targetDocument.createTextNode(s));
}

var allNodes = [];
var allElements = [];
var allRanges = [];
var allFragments = [];

function addElements(n) {

  allElements.push(n);

  var i,c;
  for (i=0;c=n.childNodes[i];++i) {
    try {
      allNodes.push(c);
      if (c.nodeType == Node.ELEMENT_NODE)
        addElements(c);
    } catch(e) { /* work around bug 335976 */ }
  }
}






function initFuzzerSpecific()
{
  dumpln("initFuzzerSpecific");
  // Create an html:iframe in HTML mode (so designMode can be used 320092)
  targetIframe = doc.createElementNS('http://www.w3.org/1999/xhtml', 'iframe');
  targetIframe.src = "data:text/html,";
  targetIframe.setAttribute("style", "width: 700px; height: 500px; border: 1px dotted green;");
  (doc.body || doc.documentElement).appendChild(targetIframe);
  
  setTimeout(init2, 1000); // or use iframe onload?
}


function init2()
{
  dumpln("init2");
  targetWindow = targetIframe.contentWindow;
  targetDocument = targetWindow.document;
  
  // Stuff everything (except top-level scripts) from our body into it.
  
  var myRoot = doc.body || doc.getElementsByTagName("body")[0] || doc.documentElement;
  for (var i = 0; i < myRoot.childNodes.length; ++i) {
    var node = myRoot.childNodes[i];
    if (!(node.tagName && node.tagName.toLowerCase().indexOf("script") != -1) && node != targetIframe) {
      // Really want adoptNode, but importNode is more widely supported.

      // Add a copy to targetDocument.
      targetDocument.body.appendChild(targetDocument.importNode(node, true));
      
      // Remove the original from this document.
      node.parentNode.removeChild(node);
      
      --i;
    }
  }
  
  try {
    targetDocument.designMode = 'on'; // perhaps it should toggle this every once in a while?
    setTimeout(init3, 1000);
  } catch(e) {
    dumpln("Can't set designMode to 'on':")
    dumpln(e);
    gPageCompleted = true;
  }
  
}


function init3()
{
  dumpln("init3");
  var myRoot = targetDocument.getElementsByTagName("body")[0] || targetDocument.documentElement; 
  addElements(myRoot);

  startFuzzing();
}
