// === Paste output from first run here, replacing the following two lines: ===

var fuzzSettings = null;
var fuzzCommands = null;

// === End pasteover ===


var fuzzerName = "Stir Table 3.1";
var recordStrategy = "Record as it goes"; // this fuzzer makes lots of new elements and stuff, so...
var suggestedMaxSteps = 1500; // Stir Table is painful to reduce, and it's pretty active


/*

Data-driven version of Stir Table

Goals:
* As little duplicated code as possible.
* Flexibility: should be able to create several tr's "at once", for example, which probably requires not using the variable name "tr" for each tr.
* Smaller statements
  * "compound commands", queued
    * "append to document first" -- don't have to use global variables for the things that aren't in the document, can continue using document.getElementById.  Can use elementsWantingChildren
    * "append to document last" -- more similar to what we had before

  pendingCommands or elementsWantingChildren.

*/

// Subset of Random Classes style properties.
var CSSProperties = [
  {
    prop: "display",
    values: ["inline", "block", "inline-block", "inline-table", "-moz-inline-block", "list-item",
    // Table
    "table", "table-row-group", "table-header-group", "table-footer-group", "table-row", "table-column-group",
    "table-column", "table-cell", "table-caption"
    ]
  },
  {
    prop: "float",
    values: ["left", "right", "none"]
  },
  {
    prop: "borderCollapse",
    values: ["inherit", "separate", "collapse"]
  },
  {
    prop: "captionSide",
    values: ["inherit", "right", "left", "top", "bottom"]
  },
  {
    prop: "emptyCells",
    values: ["inherit", "show"]
  },
  {
    prop: "tableLayout",
    values: ["fixed", "auto"]
  },
  {
    prop: "visibility",
    values: ["visible", "hidden", "collapse"]
  },
  {
    prop: "position",
    values: ["static", "relative", "absolute", "fixed"]
  },
  {
    prop: "overflow",
    values: ["visible", "scroll", "hidden", "auto", "-moz-scrollbars-horizontal", "-moz-scrollbars-none", "-moz-scrollbars-vertical", "-moz-hidden-unscrollable"]
  },
  {
    prop: "direction",
    values: ["inherit", "ltr", "rtl"]
  },
  {
    prop: "border",
    values: ["1px solid green", "2px dotted yellow", "3px double red", "6px inset green", "7px solid pink", "none"]
  },
  {
    prop: "borderLeft",
    values: ["1px solid green", "2px dotted yellow", "3px double red", "6px inset green", "7px solid pink", "none"]
  },
  {
    prop: "borderRight",
    values: ["1px solid green", "2px dotted yellow", "3px double red", "6px inset green", "7px solid pink", "none"]
  },
  {
    prop: "borderBottom",
    values: ["1px solid green", "2px dotted yellow", "3px double red", "6px inset green", "7px solid pink", "none"]
  },
  {
    prop: "borderTop",
    values: ["1px solid green", "2px dotted yellow", "3px double red", "6px inset green", "7px solid pink", "none"]
  },
  {
    prop: "outline",
    values: ["7px dotted pink", "none"]
  }

];


var lengthUnits = ['','%','px','em'];
var lengthValues = ['0','1','13','50','99','100'];

function rndWHV(elemVar, elem, tag) {
  var command = "";
  
  if (rnd(3) == 0)
    command += elemVar + ".setAttribute('width', '"  + rndElt(lengthValues) + rndElt(lengthUnits) + "'); ";
  
  if (rnd(3) == 0)
    command += elemVar + ".setAttribute('height', '" + rndElt(lengthValues) + rndElt(lengthUnits) + "'); ";
  
  while (rnd(3) == 0)
    command += change_style(elemVar, elem, tag);
  
  return command;
}
  
// highestId is a oounter for generated IDs.
// Starts at 3 because initially, the default Stir Table document contains things like table2 and td2.
// Note that generated numbers are not repeated -- there won't be a tbody5 and a td5.
highestId = 3;

// Ensure everything we care about has an ID.
function ensureIDs()
{
  var i, elems, j, elem;
  for (i = 0; i < tableRelatedTags.length; ++i) {
    elems = doc.getElementsByTagName(tableRelatedTags[i]);
    for (j = 0; elem = elems[j]; ++j)
      if (!elem.getAttribute("id"))
        bestowId(elem);
  }
}

// Used only by ensureIDs.
function bestowId(elem) {
  ++highestId;
  elem.setAttribute("id", elem.tagName.toLowerCase() + highestId);
}

// Used everywhere else.
function makeBestowId(elemVar, tag) {
  ++highestId;
  return elemVar + ".setAttribute('id', '" + tag.toLowerCase() + highestId + "'); ";
}


var elementsWantingChildren = [];

// This doesn't include rowspan/colspan/span -- those are handled separately.
var attributeInfo = {
  align: [
    "top", "bottom", "left", "right", // for caption
    "left", "right", "center", "justify", "char" // for most other things
    ],
  dir: ["ltr", "rtl"],
  width: [
    "0*", "100px", "50px 50px 50px", "30% 30% 30%", "30% 30% 30% 30%", // for colgroup and col
    "0", "4", "100", "1000", "20%", "100%" // for other things
    ],
  frame: ["void", "above", "below", "hsides", "lhs", "rhs", "vsides", "box", "border"],
  rules: ["none", "groups", "rows", "cols", "all"],
  border: ["0", "1", "2", "10"],
  cellspacing: ["0", "1", "5", "6"],
  cellpadding: ["0", "1", "5", "6"]
};

var allAttributes = [];
for (attr in attributeInfo)
  allAttributes.push(attr);

var tagInfo =
{
  td:       { spans: true,  children: ["text", "div", "span", "table"] },
  th:       { spans: true,  children: ["text", "div", "span", "table"] },

  tr:       { spans: false, children: ["td", "th"] },

  thead:    { spans: false, children: ["tr"] },
  tbody:    { spans: false, children: ["tr"] },
  tfoot:    { spans: false, children: ["tr"] },
  
  caption:  { spans: false, children: ["text", "div", "span", "table"], attributes: ["align"] },
  
  // I'm not sure about these, especially the "spans" part...
  // "span" attribute instead of "colspan" attribute?
  colgroup: { spans: true,  children: ["col"], attributes: ["width"] },
  col:      { spans: true,  children: [], attributes: ["width"] },
  
  table:    { spans: false, children: ["tr", "tr", "tr", "thead", "tbody", "tfoot", "caption", "colgroup"], attributes: ["dir", "frame", "border", "cellspacing", "cellpadding", "width"] },
  
  div:      { spans: false, children: ["text", "div", "span", "table"] },
  span:     { spans: false, children: ["text", "div", "span", "table"] } // span-containing-div causes special code to execute, so it definitely shouldn't be discouraged  
}

var tableRelatedTags = [];
for (var tagX in tagInfo)
  tableRelatedTags.push(tagX);
  

var HTML_NS = "http://www.w3.org/1999/xhtml";

function makeNewChildFor(parentTag)
{
  // Usually choose a tag based on the list of "likely child tags" for tagInfo, but sometimes choose a tag completely at random.
  var preferredChildTags = tagInfo[parentTag].children;
  var childTag;
  if (preferredChildTags.length != 0 && rnd(5) != 1)
    childTag = rndElt(preferredChildTags);
  else
    childTag = rndElt(tableRelatedTags);

  if (childTag == "table") {
    // Don't make nested tables too often.
    if (rnd(10) != 1)
      childTag = "text";
  }

  // Special-case text nodes.  
  if (childTag == "text")
    return "var newNode = document.createTextNode('" + parentTag + "'); ";
    
  var command = "var newNode = document.createElementNS(HTML_NS, '" + childTag + "'); ";
  
  if (tagInfo[childTag].children[0] == "text" && rnd(5) != 1)
    command += "newNode.appendChild(document.createTextNode('" + childTag.toUpperCase() + "')); ";
  
  command += makeBestowId("newNode", childTag);
  
  command += rndWHV("newNode", null, childTag);

  if (tagInfo[childTag].spans) {
    if (rnd(3) == 0)
      command += change_rowspan("newNode", null, childTag);
    if (rnd(3) == 0)
      command += change_colspan("newNode", null, childTag);  
  }
    
  return command;
}

function appendNewInto(elemVar, elem, tag) {
  var command = makeNewChildFor(tag); // Always declares and creates something called "newNode".
  command += elemVar + ".appendChild(newNode); bless(newNode);";
  return command;
}

function getElementByIdExpr(id) { return "doc.getElementById('" + id + "')"; } 

function insertNewInto(elemVar, elem, tag)
{
  var sib = elem.firstChild ? rndElt(elem.childNodes) : null;

  // Ignore text nodes and other nodes without IDs.  This isn't quite right :( --jtr
  while (sib && (!sib.getAttribute || !sib.getAttribute("id")))
    sib = sib.nextSibling;
    
  var sibExpr = sib ? 
    getElementByIdExpr(sib.getAttribute("id")) :
    "null";
    
  var command = makeNewChildFor(tag); // Always declares and creates something called "newNode".
  command += elemVar + ".insertBefore(newNode, " + sibExpr + "); bless(newNode);";
  return command;
}

// This strange function is called *by* randomly generated functions, and is used only to *influence* future randomly generated functions.
function bless(newNode)
{
  if (newNode.nodeType != Node.ELEMENT_NODE)
    return;

  var numChildrenWanted;
  
  if (tagInfo[newNode.tagName.toLowerCase()].children.length)
    numChildrenWanted = rndElt([0, 0, 1, 1, 1, 1, 2]);
  else
    numChildrenWanted = rndElt([0, 0, 0, 0, 0, 0, 1]);
  
  for (var i = 0; i < numChildrenWanted; ++i)
    elementsWantingChildren.push(newNode);
}

    




function remove(elemVar, elem, tag) {
  // Never remove tables or divs (???)
  if (tag == "table" || tag == "div")
    return;
     
  return elemVar + ".parentNode.removeChild(" + elemVar + "); ";
}

function change_style(elemVar, elem, tag) {
  var propPair = rndElt(CSSProperties);
  var prop = propPair.prop;
  var value;
  
  if (rnd(3) == 1)
    value = rndElt(propPair.values);
  else
    value = "";
  
  return elemVar + ".style." + prop + " = \"" + value + "\"; ";
  
  /*

  if (elem && elem.style && elem.style.visibility == "collapse") {
    return elemVar + ".style.visibility = 'visible'; ";
  } else {
    if ((tag == "table" || tag == "div") && rnd(90) != 3) {
      // Avoid making it boring.
      return "";
    }
    
    if (rnd(4) == 1)
      return elemVar + ".style.visibility = 'collapse'; ";
  }
  
  return "";
  
  */
}

function change_attribute(elemVar, elem, tag) {
  var attrs = tagInfo[tag].attributes;
  if (!attrs || rnd(5) == 1)
    attrs = allAttributes;
  var attr = rndElt(attrs);
  
  if (rnd(3)) {
    var value = rndElt(attributeInfo[attr]);
    return "/* CASA */ " + elemVar + ".setAttribute('" + attr + "', '" + value + "');";
  } else {
    return "/* CARA */ " + elemVar + ".removeAttribute('" + attr + "');";
  }
}

function change_rowspan(elemVar, elem, tag) {
  var rowspanX = rnd(4);
  return elemVar + ".setAttribute('rowspan', " + rowspanX + "); ";
}

function change_colspan(elemVar, elem, tag) {
  var colspanX = rnd(4);
  
  if (tag.toLowerCase() == 'col' || tag.toLowerCase() == 'colgroup')
    return elemVar + ".setAttribute('span', " + colspanX + "); ";
  else
    return elemVar + ".setAttribute('colspan', " + colspanX + "); ";
}


// ops are functions: (elemVar, possibly null elem, tag) -> command string
var ops = [insertNewInto, appendNewInto, remove, change_rowspan, change_colspan, change_style, change_attribute];


function randomVictim()
{
  while (1) {
    var tag = rndElt(tableRelatedTags);
    var elems = document.getElementsByTagName(tag);
    if (elems.length) {
      elem = rndElt(document.getElementsByTagName(tag));
      if(elem.getAttribute("id"))
        return elem;
      else
        dumpln("Table-related element with no id?");
    }
  }
}


function makeCommand() {

  if (doc.getElementsByTagName("table").length == 0) {
    dumpln("No tables!  I'm done!")
    gPageCompleted = true; // tell spider we're not interested in this page (any more?)
    return ""; // don't get stuck in an infinite loop (we may get called again, though!?)
  }
  
  var command = "";
    
  var elem, op;

  if (elementsWantingChildren.length) {
    // Pop off an element, and make a child for it.
    elem = elementsWantingChildren.pop();
    op = appendNewInto;
  } else if (rnd(20) == 6) {
    // Random appendChild (like in Stir DOM)
    elem1 = randomVictim();
    elem2 = randomVictim();
    elem1Id = elem1.getAttribute("id");
    elem2Id = elem2.getAttribute("id");
    return getElementByIdExpr(elem1Id) + ".appendChild(" + getElementByIdExpr(elem2Id) + "); ";
  } else {
    // Pick an operation and a victim at random.
    op = rndElt(ops);
    elem = randomVictim();
  }
  
  var tag = elem.tagName.toLowerCase();

  var elemId = elem.getAttribute("id");
  var elemVar = "$" + elemId.replace(/\-/g, "_"); // :)
  
  command += "var " + elemVar + " = " + getElementByIdExpr(elemId) + "; ";
  command += op(elemVar, elem, tag); // pass lots of seemingly redundant information: a variable name for the element, the element, and its tagName!!
  
  return command;
}

function initFuzzerSpecific()
{
  ensureIDs();    
  startFuzzing();
}

